# gozerbot/runner.py
#
#

""" thread to run jobs """

__copyright__ = 'this file is in the public domain'

from gozerbot.thr import start_new_thread
from gozerbot.generic import handle_exception, rlog
import Queue, time

class Runner(object):

    def __init__(self):
        self.queue = Queue.Queue()
        self.nr = 0
        self.dostop = False
        self.running = False

    def put(self, descr, func, *args, **kwargs):
        self.queue.put_nowait((descr, func, args, kwargs))

    def start(self):
        if not self.running:
            start_new_thread(self.run, ())
            rlog(1, 'runner', 'started runner thread')

    def stop(self):
        self.dostop = True
        self.running = False
        self.queue.put_nowait((None, None, None, None))

    def run(self):
        self.running = True
        while not self.dostop:
            (descr, func, args, kwargs) = self.queue.get()
            if self.dostop:
                break
            try:
                rlog(1, 'runner', 'running %s: %s' % (descr, str(func)))
                start = time.time()
                func(*args, **kwargs)
                elapsed = time.time() - start
                if elapsed > 1:
                    rlog(10, 'runner', 'ALERT %s %s job taking too long: %s \
seconds' % (descr, str(func), elapsed))
            except Exception, ex:
                handle_exception()
        rlog(10, 'runner', 'stopped runner thread')

runner = Runner()
